
/* file p7ck.c
 created lun. déc. 13 07:31:56 CET 2021 by Yann Guidon whygee@f-cpu.org

 definition of the format that packs 7 bits per byte

 gcc -Os -S p7ck.c
 gcc -Wall -DTEST -Os p7ck.c
*/

#include <stdint.h>
#include <inttypes.h>


typedef struct p7k_struct {
  uint8_t Xorig[5];
  uint8_t Xend[5];
  uint8_t arcs[5];
  uint8_t arclen[10];
  uint8_t z;  // 26th byte
} p7k_struct;

void encode_p7k(uint32_t Xorig, uint32_t Xend, uint32_t arcs,
 uint64_t arclen, p7k_struct *p7k) {
  void encode_35(uint32_t dat, uint8_t *c) {
    c[4]= 128 |  dat;
    c[3]= 128 | (dat >>  7);
    c[2]= 128 | (dat >> 14);
    c[1]= 128 | (dat >> 21);
    c[0]= 128 | (dat >> 28);
  }

  void encode_70(uint64_t dat, uint8_t *c) {
    c[9]= 128 |  dat;
    c[8]= 128 | (dat >>  7);
    c[7]= 128 | (dat >> 14);
    c[6]= 128 | (dat >> 21);
    c[5]= 128 | (dat >> 28);
    c[4]= 128 | (dat >> 35);
    c[3]= 128 | (dat >> 42);
    c[2]= 128 | (dat >> 49);
    c[1]= 128 | (dat >> 56);
    c[0]= 128 | (dat >> 63);
  }

  encode_35(Xorig , &p7k->Xorig[0] );
  encode_35(Xend  , &p7k->Xend[0]  );
  encode_35(arcs  , &p7k->arcs[0]  );
  encode_70(arclen, &p7k->arclen[0]);
  p7k->z = 0;
}

void decode_p7k(uint32_t *Xorig, uint32_t *Xend, uint32_t *arcs,
   uint64_t *arclen, p7k_struct *p7k) {

  void decode_35(uint32_t *dat, uint8_t *c) {
    *(uint32_t *) c &= 0x7F7F7F7FUL;

    uint32_t t = c[0] << 28;
    uint32_t u = c[1] << 21;
    uint32_t v = c[2] << 14;
    uint32_t x = c[3] << 7;
    *dat = (c[4] & 127) | t | u | v | x;
  }

  void decode_70(uint64_t *dat, uint8_t *c) {
    *(uint64_t *) c &= 0x7F7F7F7F7F7F7F7FULL;

    uint64_t e = c[0];       e <<= 63;
    uint64_t f = c[1];       f <<= 56;
    uint64_t g = c[2];       g <<= 49;
    uint64_t h = c[3];       h <<= 42;
    uint64_t i = c[4];       i <<= 35;
    uint64_t j = c[5];       j <<= 28;
    uint64_t k = c[6];       k <<= 21;
    uint64_t l = c[7];       l <<= 14;
    uint64_t m = c[8] & 127; m <<= 7;
    *dat = (c[9] & 127)
      |e|f|g|h|i|j|k|l|m;
  }

  decode_35(Xorig , &p7k->Xorig[0] );
  decode_35(Xend  , &p7k->Xend[0]  );
  decode_35(arcs  , &p7k->arcs[0]  );
  decode_70(arclen, &p7k->arclen[0]);
}

#ifdef TEST
#include <stdio.h>
#include <stdlib.h>

// Générateur de nombres pseudo-aléatoires
// PEAC16 : période de 2 milliards
///////// PRNG :
uint16_t PRNGX=0x1234,
         PRNGY=0x5678;
uint32_t PRNGC=0;

uint16_t PRNG() {
  PRNGC += (uint32_t)PRNGX;
  PRNGC += (uint32_t)PRNGY;
  PRNGY = PRNGX;
  PRNGX = (uint16_t) PRNGC;
  PRNGC >>= 16;
  return PRNGX;
}

int main(int argc, char *argv[]) {
  p7k_struct p;
  uint32_t a, b, c;
  uint64_t d;

  for (int i=0; i<10000; i++) {
    uint32_t e = PRNG() | (PRNG() << 16);
    uint32_t f = PRNG() | (PRNG() << 16);
    uint32_t g = PRNG() | (PRNG() << 16);
    uint64_t h = PRNG() | (PRNG() << 16)
             |  ((uint64_t)PRNG() << 32)
             |  ((uint64_t)PRNG() << 48);
    encode_p7k( e, f, g, h, &p);
    decode_p7k( &a, &b, &c, &d, &p);
//    printf("%u %u %u %"PRIu64"\n", a, b, c, d);
    if (( a!=e )
      ||( b!=f )
      ||( c!=g )
      ||( d!=h )
    ) {
      printf(" mismatch !!\n");
      return EXIT_FAILURE;
    }
  }
  return EXIT_SUCCESS;
}
#endif
