
/*  File scan_1x_06.c
 created ven. nov. 26 21:30:47 CET 2021
 version ven. nov. 26 21:43:33 CET 2021 now with unrolled loop counter : close to no difference
 version sam. nov. 27 22:33:51 CET 2021 : 2 conditions per loop phase
 version ven. déc.  3 20:01:12 CET 2021 : adapted from faster_02.c
 version sam. déc.  4 20:31:50 CET 2021 : adapted from ../p4ck/test_w16_02.c
 version mar. déc.  7 01:41:29 CET 2021 : adding the arc count
 version lun. déc. 13 09:24:20 CET 2021 : adding p7k
 version lun. déc. 13 21:39:34 CET 2021 : changed the command line format, Xstart to Xlim+1

 compile & run :
   gcc -Wall -Os -DWIDTH=16 scan_1x_05.c -o scan && ./scan

*/

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <inttypes.h>
#include <errno.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#ifndef WIDTH
#define WIDTH (16)
#endif

#define SIZE  (1<<WIDTH)
#define TOTAL ((((SIZE+1ULL)*SIZE)>>1)-2ULL)
#define MASK  (SIZE-1)


#include "p7ck.c"
#ifndef BUFFZIZE
#define BUFFZIZE (16384)
#endif
p7k_struct results[BUFFZIZE];

void semiarc(uint64_t Xorigin, uint64_t *len, uint64_t *Xcross) {
  uint64_t arclen=0;
  uint32_t
    X = Xorigin,
    Y = 0,
    C = 0;

loopstart:
      C += X + Y;
      Y = C & MASK;  C = C >> WIDTH;
      if (X==0)
        goto endofloopodd;
      if (X==MASK)
        goto endofloopodd_comp;

      C += X + Y;
      arclen+=2;
      X = C & MASK;  C = C >> WIDTH;
      if (Y==MASK)
        goto endofloopeven_comp;
      if (Y!=0)
        goto loopstart;

//endofloopeven:
//  putchar('E');
  *Xcross = X;
  *len = arclen;
  return;

endofloopeven_comp:
//  putchar('e');
  *Xcross = MASK-X;
  *len = arclen;
  return;

endofloopodd:
//  putchar('O');
  *Xcross = Y;
  *len = arclen+1;
  return;

endofloopodd_comp:
//  putchar('o');
  *Xcross = MASK-Y;
  *len = arclen+1;
  return;
}


int main(int argc, char *argv[]) {
  uint64_t
    X2,
    len=0,
    total=0;
  int64_t Xmax=SIZE, X=1;
  int fout, res_index;
  char *c, *outname="log.p7k";

// Parse the command line arguments
  if (argc != 1) {
    if (argc == 4) {
      X=strtoul(argv[1], &c, 10);
      if (errno || *c!=0 || X > MASK || X < 0) {
        printf("invalid X parameter : %s ??\n", argv[1]);
        return EXIT_FAILURE;
      }
      Xmax=strtoul(argv[2], &c, 10);
      if (errno || *c!=0 || Xmax > SIZE || Xmax < X) {
        printf("invalid count parameter : %s ??\n", argv[2]);
        return EXIT_FAILURE;
      }
      outname=argv[3];
    }
    else {
      printf("Usage : %s (origin X) (number of scans) (output file name)\nWidth=%d bits\n", argv[0], WIDTH);
      return EXIT_FAILURE;
    }
  }
  printf("Scanning w%d from X=%"PRId64" (included) to %"PRId64" (not included)\n", WIDTH, X, Xmax);

  if ((fout=open(outname, O_CREAT|O_WRONLY|O_TRUNC, O_RDWR)) == -1) {
    perror("can't open output file for writing");
    return EXIT_FAILURE;
  }

// Loop body:
  res_index=0;
  do {
    if (X==0) {
      X2=1;
      len=1;
    }
    else
      semiarc(X, &len, &X2);
//    printf("%"PRId64" %"PRIu64" %"PRIu64" 1\n", X2, X, len);
    encode_p7k(X, X2, 1,  len, &results[res_index]);

    res_index++;
    if (res_index >= BUFFZIZE) {
      printf("%"PRId64"\n", X);
      write(fout, &results, sizeof results);
      res_index=0;
    }

    total += len;
    X++;
  } while (X < Xmax);

  if (res_index > 0) {
    write(fout, &results, res_index * sizeof results[0]);
    printf("%"PRId64"\n", X);
  }
  printf(" %s done.\n", outname);
  return EXIT_SUCCESS;
}
